from dataclasses import dataclass
from typing import Any, Union
from math import sqrt
from sympy import symbols, nsimplify, integrate
from sympy.core.mul import Mul

import rcdesign.is456 as is456

# from rcdesign.stressblock import StressBlock


@dataclass
class LSMStressBlock:
    label: str = "IS 456 LSM"
    ecy: float = is456.ecy
    ecu: float = is456.ecu

    def __repr__(self):
        s = self.label
        return s

    def isvalid_ecmax(self, ecmax: float) -> float:
        if (ecmax < 0) or (ecmax > self.ecu):
            raise ValueError
        else:
            return ecmax

    def isvalid_k(self, k: float) -> float:
        if k < 0:
            raise ValueError
        else:
            return k

    def isvalid_z(self, z: float, k: float):
        k = self.isvalid_k(k)
        if (z < k - 1) or (z > k):
            raise ValueError
        else:
            return z

    def _ec(self, _k: float, ecmax: float = ecu) -> Union[Mul, Any]:
        ecmax = self.isvalid_ecmax(ecmax)
        z, k = symbols("z k")
        if _k < 0:  # Invalid values for k
            raise ValueError
        elif _k == 0:  # Unstressed condition
            ec = nsimplify(0)
        elif _k <= 1:  # Flexure or axial compression with NA within the section
            ec = (z / nsimplify(self.ecy / ecmax * k)).evalf(subs={"k": _k})
        else:  # Axial compression with NA outside the section
            ec = (z / nsimplify((k - 3 / 7))).evalf(subs={"k": _k})
        return ec

    def ec(self, z: float, k: float, ecmax: float = ecu) -> float:
        _ec = self._ec(k, ecmax)
        return float(_ec.evalf(subs={"z": z}))

    def _fc_(self, ec: float) -> float:
        if (ec < 0) or (ec > self.ecu):
            return 0.0
        ec_ecy = ec / self.ecy
        if ec_ecy < 1:
            return 2 * ec_ecy - ec_ecy**2
        else:
            return 1.0

    def _fc(self, z: float, k: float, ecmax: float = ecu) -> Union[Mul, Any]:
        ecmax = self.isvalid_ecmax(ecmax)
        if k < 0:  # Invalid values for k
            raise ValueError
        elif k == 0:  # Unstressed condition
            fc = nsimplify(0)
        else:  # Flexure or axial compression with NA within the section
            ec = self.ec(z, k, ecmax)
            if ec < 0:
                fc = nsimplify(0)
            elif ec < 1:
                ec_ecy = self._ec(k, ecmax)
                fc = (2 * ec_ecy - ec_ecy**2).evalf(subs={"k": k})
            else:
                fc = nsimplify(1)
        return fc

    def fc(self, z: float, k: float, ecmax: float = ecu) -> float:
        fc = self._fc(z, k, ecmax)
        return float(fc.evalf(subs={"z": z}))

    def C(
        self, z1: float, z2: float, k: float, ecmax: float = ecu
    ) -> Union[float, Any]:
        ecmax = self.isvalid_ecmax(ecmax)
        z = symbols("z")
        k = self.isvalid_k(k)
        if k == 0:
            return 0.0

        z1 = self.isvalid_z(z1, k)
        z2 = self.isvalid_z(z2, k)
        if z1 > z2:
            z1, z2 = z2, z1
        ec1 = self.ec(z1, k)
        ec2 = self.ec(z2, k)
        if ec2 <= 1:  # Parabolic only
            fcp = self._fc(z1, k, ecmax)
            Cp = integrate(fcp, (z, z1, z2))
            Cr = 0.0
        elif ec1 >= 1:  # Rectangular only
            Cp = 0.0
            fcr = self._fc(z1, k, ecmax)
            Cr = integrate(fcr, (z, z1, z2))
        else:  # Both Parabolic and Rectangular
            if k <= 1:  # NA within the section
                zz = self.ecy / ecmax * k
            else:  # NA outside the section
                zz = k - (1 - self.ecy / self.ecu)
            fcp = self._fc(z1, k, ecmax)
            fcr = self._fc(z2, k, ecmax)
            Cp = integrate(fcp, (z, z1, zz))
            Cr = integrate(fcr, (z, zz, z2))
        return Cp + Cr  # type: ignore

    def M(
        self, z1: float, z2: float, k: float, ecmax: float = ecu
    ) -> Union[float, Any]:
        ecmax = self.isvalid_ecmax(ecmax)
        z = symbols("z")
        k = self.isvalid_k(k)
        if k == 0:
            return 0.0

        z1 = self.isvalid_z(z1, k)
        z2 = self.isvalid_z(z2, k)
        if z1 > z2:
            z1, z2 = z2, z1
        ec1 = self.ec(z1, k)
        ec2 = self.ec(z2, k)
        if ec2 <= 1:  # Parabolic only
            fcp = self._fc(z1, k, ecmax)
            Mp = integrate(fcp * z, (z, z1, z2))
            Mr = 0.0
        elif ec1 >= 1:  # Rectangular only
            Mp = 0.0
            fcr = self._fc(z1, k, ecmax)
            Mr = integrate(fcr * z, (z, z1, z2))
        else:  # Both Parabolic and Rectangular
            if k <= 1:  # NA within the section
                zz = self.ecy / ecmax * k
            else:  # NA outside the section
                zz = k - (1 - self.ecy / self.ecu)
            fcp = self._fc(z1, k, ecmax)
            fcr = self._fc(z2, k, ecmax)
            Mp = integrate(fcp * z, (z, z1, zz))
            Mr = integrate(fcr * z, (z, zz, z2))
        return Mp + Mr  # type: ignore


@dataclass
class WSMStressBlock:
    _fcbc: float
    _fst: float

    @property
    def fcbc(self) -> float:
        return self._fcbc

    @fcbc.setter
    def fcbc(self, _fcbc: float) -> None:
        self._fcbc = _fcbc

    @property
    def fst(self) -> float:
        return self._fst

    @fst.setter
    def fst(self, _fst: float) -> None:
        self._fst = _fst

    @property
    def m(self) -> float:
        return 280.0 / (3 * self.fcbc)

    def kb(self, d: float = 1.0) -> float:
        m = self.m
        return (m * self.fcbc) / (m * self.fcbc + self.fst) * d

    def jb(self, d: float = 1.0) -> float:
        return d * (1 - self.kb(d) / 3.0)

    def Qb(self, b: float = 1.0, d: float = 1.0) -> float:
        xb = self.kb(d)
        # print(b, xb, fcbc, d)
        return b * xb * self.fcbc / 2.0 * (d - xb / 3.0)

    def xb(self, d):
        m = self.m
        return m * self.fcbc * d / (self.fst + m * self.fcbc)

    def x(self, b: float, d: float, Ast: float) -> float:
        m = self.m
        ca = b / 2.0
        cb = m * Ast
        cc = -m * Ast * d
        return (-cb + sqrt(cb**2 - 4 * ca * cc)) / (2 * ca)

    def reqd_Asc_Ast(self, b, d, dd: float, M):
        Mb = self.Qb(b, d)
        if M <= Mb:
            x = (1.5 - sqrt(1.5**2 - (6 * M / (self.fcbc * b * d**2)))) * d
            Ast = b * x * self.fcbc / (2 * self.fst)
            Asc = 0.0
            # print('Singly', x, Ast, Asc, M/1e6, Mb/1e6)
        else:
            xb = self.kb(d)
            Ast1 = b * xb * self.fcbc / (2 * self.fst)
            M2 = M - Mb
            Ast2 = M2 / (self.fst * (d - dd))
            Ast = Ast1 + Ast2
            m = self.m
            Asc = (m * Ast2 * (d - xb)) / ((1.5 * m - 1) * (xb - dd))
            # print('Doubly', xb, Asc, Ast, M/1e6, Mb/1e6)
        return Asc, Ast



from fractions import Fraction
import numpy as np
import numpy.typing as ntp


gamma_c: Fraction = Fraction(3, 2)  # 1.5
gamma_s: Fraction = Fraction(115, 100)  # 1.15
ecy: float = 0.002
ecu: float = 0.0035
Es: float = 2e5
fdc: Fraction = Fraction(2, 3) / gamma_c
fds: Fraction = 1 / gamma_s
_A1: Fraction = Fraction(2, 3) * Fraction(4, 7)
_A2: Fraction = Fraction(3, 7)
_A = _A1 + _A2
k1: Fraction = _A * fdc
k2: Fraction = 1 - (_A1 * (Fraction(5, 8) * Fraction(4, 7)) + _A2 * (Fraction(4, 7) + Fraction(1, 2) * Fraction(3, 7))) / _A

inel_strain: ntp.NDArray = np.array(
    [[0.8, 0.0], [0.85, 0.0001], [0.9, 0.0003], [0.95, 0.0007], [0.975, 0.001], [1, 0.002]]
)



from dataclasses import dataclass
from enum import Enum
from rcdesign.is456.constants import fds


class Exposure(Enum):
    MILD = 1
    MODERATE = 2
    SEVERE = 3
    VERY_SEVERE = 4
    EXTREME = 5


@dataclass
class Beam:
    def min_ver_spacing(
        self, max_bardia: float, nominal_agg_size: float = 20.0
    ) -> float:
        return max(15.0, 2 / 3 * nominal_agg_size, max_bardia)

    def min_hor_spacing(
        self, max_bardia: float, nominal_agg_size: float = 20.0
    ) -> float:
        return max(max_bardia, nominal_agg_size + 5.0)

    def Ast_min(self, b: float, d: float, fy: float) -> float:
        return 0.85 / fy * b * d

    def Ast_max(self, b: float, D: float) -> float:
        return 0.04 * b * D

    def Asc_max(self, b: float, D: float) -> float:
        return 0.04 * b * D

    def Ast_side_face(self, bw: float, dw: float) -> float:
        if dw < 750:
            return 0.0
        else:
            return 0.1 / 100 * bw * dw

    def sv_max(self, d: float, alpha: float = 90.0) -> float:
        return min(300.0, 0.75 * d if alpha == 90 else d)

    def sv_min(self, b: float, fy: float, Asv: float) -> float:
        return Asv * fds * min(415, fy) / (0.4 * b)

    def nominal_cover_durability(self, exposure: Exposure, bardia: float, fck: float) -> float:
        match exposure:
            case Exposure.MILD:
                cover = 20 if bardia > 12 else 15
            case Exposure.MODERATE:
                cover = 30
            case Exposure.SEVERE:
                cover = 45 if fck < 35 else 40
            case Exposure.VERY_SEVERE:
                cover = 50 if fck < 35 else 45
            case Exposure.EXTREME:
                cover = 75
            case _:
                raise AttributeError
        return float(cover)

    def nominal_cover_fire(self) -> float:
        raise NotImplementedError


@dataclass
class Slab:
    def Ast_min(self, D: float, fy: float) -> float:
        pst = 0.15 if fy < 415 else 0.12
        return pst / 100 * 1000 * D

    def Ast_max(self, D: float) -> float:
        return Beam().Ast_max(1000.0, D)

    def bardia_max(self, D: float) -> float:
        return D / 8.0

    def max_hor_spacing(self, d: float, main_bars: bool = True) -> float:
        if main_bars:
            return max(3 * d, 300.0)
        else:
            return max(5 * d, 450.0)

    def nominal_cover_durability(self, exposure: Exposure, bardia: float, fck: float) -> float:
        return Beam().nominal_cover_durability(exposure, bardia, fck)

    def nominal_cover_fire(self) -> float:
        raise NotImplementedError


@dataclass
class Column:
    def Asc_min(self, b: float, D: float) -> float:
        return 0.8 / 100 * b * D

    def Asc_max(self, b: float, D: float) -> float:
        return 6 / 100.0 * b * D

    def bardia_min(self) -> float:
        return 12.0

    def lateral_ties_pitch(self, b: float, bardia_min: float) -> float:
        return min(b, 16 * bardia_min, 300.0)

    def lateral_ties_dia(self, bardia_max: float) -> float:
        return max(bardia_max / 4.0, 6.0)
